@@ -1,28 +1,52 @@
+/*
+ * Copyright 2021 l1npengtul <l1npengtul@protonmail.com> / The Nokhwa Contributors
+ *
+ * Licensed under the Apache License, Version 2.0 (the "License");
+ * you may not use this file except in compliance with the License.
+ * You may obtain a copy of the License at
+ *
+ *     http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing, software
+ * distributed under the License is distributed on an "AS IS" BASIS,
+ * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+ * See the License for the specific language governing permissions and
+ * limitations under the License.
+ */
+
 use crate::{
     CameraFormat, CameraInfo, CaptureAPIBackend, CaptureBackendTrait, FrameFormat, NokhwaError,
     Resolution,
 };
 use image::{buffer::ConvertBuffer, ImageBuffer, Rgb, RgbaImage};
-use std::{cell::RefCell, collections::HashMap};
+use std::{borrow::Cow, collections::HashMap};
 #[cfg(feature = "output-wgpu")]
 use wgpu::{
     Device as WgpuDevice, Extent3d, ImageCopyTexture, ImageDataLayout, Queue as WgpuQueue,
-    Texture as WgpuTexture, TextureDescriptor, TextureDimension, TextureFormat, TextureUsage,
+    Texture as WgpuTexture, TextureAspect, TextureDescriptor, TextureDimension, TextureFormat,
+    TextureUsages,
 };
 
 /// The main `Camera` struct. This is the struct that abstracts over all the backends, providing a simplified interface for use.
 pub struct Camera {
     idx: usize,
-    backend: RefCell<Box<dyn CaptureBackendTrait>>,
+    backend: Box<dyn CaptureBackendTrait>,
     backend_api: CaptureAPIBackend,
 }
 
 #[allow(clippy::nonminimal_bool)]
 impl Camera {
+    /// Create a new camera from an `index` and `format`
+    /// # Errors
+    /// This will error if you either have a bad platform configuration (e.g. `input-v4l` but not on linux) or the backend cannot create the camera (e.g. permission denied).
+    pub fn new(index: usize, format: Option<CameraFormat>) -> Result<Self, NokhwaError> {
+        Camera::with_backend(index, format, CaptureAPIBackend::Auto)
+    }
+
     /// Create a new camera from an `index`, `format`, and `backend`. `format` can be `None`.
     /// # Errors
     /// This will error if you either have a bad platform configuration (e.g. `input-v4l` but not on linux) or the backend cannot create the camera (e.g. permission denied).
-    pub fn new(
+    pub fn with_backend(
         index: usize,
         format: Option<CameraFormat>,
         backend: CaptureAPIBackend,
@@ -31,7 +55,7 @@ impl Camera {
 
         Ok(Camera {
             idx: index,
-            backend: RefCell::new(camera_backend),
+            backend: camera_backend,
             backend_api: backend,
         })
     }
@@ -48,10 +72,11 @@ impl Camera {
         backend: CaptureAPIBackend,
     ) -> Result<Self, NokhwaError> {
         let camera_format = CameraFormat::new_from(width, height, fourcc, fps);
-        Camera::new(index, Some(camera_format), backend)
+        Camera::with_backend(index, Some(camera_format), backend)
     }
 
     /// Gets the current Camera's index.
+    #[must_use]
     pub fn index(&self) -> usize {
         self.idx
     }
@@ -61,15 +86,16 @@ impl Camera {
     /// The Backend may fail to initialize.
     pub fn set_index(&mut self, new_idx: usize) -> Result<(), NokhwaError> {
         {
-            self.backend.borrow_mut().stop_stream()?;
+            self.backend.stop_stream()?;
         }
-        let new_camera_format = self.backend.borrow().camera_format();
+        let new_camera_format = self.backend.camera_format();
         let new_camera = init_camera(new_idx, Some(new_camera_format), self.backend_api)?;
-        *self.backend.borrow_mut() = new_camera;
+        self.backend = new_camera;
         Ok(())
     }
 
     /// Gets the current Camera's backend
+    #[must_use]
     pub fn backend(&self) -> CaptureAPIBackend {
         self.backend_api
     }
@@ -79,105 +105,115 @@ impl Camera {
     /// The new backend may not exist or may fail to initialize the new camera.
     pub fn set_backend(&mut self, new_backend: CaptureAPIBackend) -> Result<(), NokhwaError> {
         {
-            self.backend.borrow_mut().stop_stream()?;
+            self.backend.stop_stream()?;
         }
-        let new_camera_format = self.backend.borrow().camera_format();
+        let new_camera_format = self.backend.camera_format();
         let new_camera = init_camera(self.idx, Some(new_camera_format), new_backend)?;
-        *self.backend.borrow_mut() = new_camera;
+        self.backend = new_camera;
         Ok(())
     }
 
     /// Gets the camera information such as Name and Index as a [`CameraInfo`].
-    pub fn info(&self) -> CameraInfo {
-        self.backend.borrow().camera_info()
+    #[must_use]
+    pub fn info(&self) -> &CameraInfo {
+        self.backend.camera_info()
     }
     /// Gets the current [`CameraFormat`].
+    #[must_use]
     pub fn camera_format(&self) -> CameraFormat {
-        self.backend.borrow().camera_format()
+        self.backend.camera_format()
     }
     /// Will set the current [`CameraFormat`]
     /// This will reset the current stream if used while stream is opened.
     /// # Errors
     /// If you started the stream and the camera rejects the new camera format, this will return an error.
     pub fn set_camera_format(&mut self, new_fmt: CameraFormat) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().set_camera_format(new_fmt)
+        self.backend.set_camera_format(new_fmt)
     }
     /// A hashmap of [`Resolution`]s mapped to framerates
     /// # Errors
-    /// This will error if the camera is not queryable or a query operation has failed. Some backends will error this out as a Unsupported Operation ([`NokhwaError::UnsupportedOperation`]).
+    /// This will error if the camera is not queryable or a query operation has failed. Some backends will error this out as a [`UnsupportedOperationError`](crate::NokhwaError::UnsupportedOperationError).
     pub fn compatible_list_by_resolution(
-        &self,
+        &mut self,
         fourcc: FrameFormat,
     ) -> Result<HashMap<Resolution, Vec<u32>>, NokhwaError> {
-        self.backend.borrow().compatible_list_by_resolution(fourcc)
+        self.backend.compatible_list_by_resolution(fourcc)
     }
     /// A Vector of compatible [`FrameFormat`]s.
     /// # Errors
-    /// This will error if the camera is not queryable or a query operation has failed. Some backends will error this out as a Unsupported Operation ([`NokhwaError::UnsupportedOperation`]).
-    pub fn compatible_fourcc(&self) -> Result<Vec<FrameFormat>, NokhwaError> {
-        self.backend.borrow_mut().compatible_fourcc()
+    /// This will error if the camera is not queryable or a query operation has failed. Some backends will error this out as a [`UnsupportedOperationError`](crate::NokhwaError::UnsupportedOperationError).
+    pub fn compatible_fourcc(&mut self) -> Result<Vec<FrameFormat>, NokhwaError> {
+        self.backend.compatible_fourcc()
     }
     /// Gets the current camera resolution (See: [`Resolution`], [`CameraFormat`]).
+    #[must_use]
     pub fn resolution(&self) -> Resolution {
-        self.backend.borrow().resolution()
+        self.backend.resolution()
     }
     /// Will set the current [`Resolution`]
     /// This will reset the current stream if used while stream is opened.
     /// # Errors
     /// If you started the stream and the camera rejects the new resolution, this will return an error.
     pub fn set_resolution(&mut self, new_res: Resolution) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().set_resolution(new_res)
+        self.backend.set_resolution(new_res)
     }
     /// Gets the current camera framerate (See: [`CameraFormat`]).
+    #[must_use]
     pub fn frame_rate(&self) -> u32 {
-        self.backend.borrow().frame_rate()
+        self.backend.frame_rate()
     }
     /// Will set the current framerate
     /// This will reset the current stream if used while stream is opened.
     /// # Errors
     /// If you started the stream and the camera rejects the new framerate, this will return an error.
     pub fn set_frame_rate(&mut self, new_fps: u32) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().set_frame_rate(new_fps)
+        self.backend.set_frame_rate(new_fps)
     }
     /// Gets the current camera's frame format (See: [`FrameFormat`], [`CameraFormat`]).
+    #[must_use]
     pub fn frame_format(&self) -> FrameFormat {
-        self.backend.borrow().frame_format()
+        self.backend.frame_format()
     }
     /// Will set the current [`FrameFormat`]
     /// This will reset the current stream if used while stream is opened.
     /// # Errors
     /// If you started the stream and the camera rejects the new frame format, this will return an error.
     pub fn set_frame_format(&mut self, fourcc: FrameFormat) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().set_frame_format(fourcc)
+        self.backend.set_frame_format(fourcc)
     }
     /// Will open the camera stream with set parameters. This will be called internally if you try and call [`frame()`](CaptureBackendTrait::frame()) before you call [`open_stream()`](CaptureBackendTrait::open_stream()).
     /// # Errors
     /// If the specific backend fails to open the camera (e.g. already taken, busy, doesn't exist anymore) this will error.
     pub fn open_stream(&mut self) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().open_stream()
+        self.backend.open_stream()
     }
     /// Checks if stream if open. If it is, it will return true.
+    #[must_use]
     pub fn is_stream_open(&self) -> bool {
-        self.backend.borrow().is_stream_open()
+        self.backend.is_stream_open()
     }
     /// Will get a frame from the camera as a Raw RGB image buffer. Depending on the backend, if you have not called [`open_stream()`](CaptureBackendTrait::open_stream()) before you called this,
     /// it will either return an error.
     /// # Errors
     /// If the backend fails to get the frame (e.g. already taken, busy, doesn't exist anymore), the decoding fails (e.g. MJPEG -> u8), or [`open_stream()`](CaptureBackendTrait::open_stream()) has not been called yet,
     /// this will error.
-    pub fn frame(&self) -> Result<ImageBuffer<Rgb<u8>, Vec<u8>>, NokhwaError> {
-        self.backend.borrow_mut().frame()
+    pub fn frame(&mut self) -> Result<ImageBuffer<Rgb<u8>, Vec<u8>>, NokhwaError> {
+        self.backend.frame()
     }
     /// Will get a frame from the camera **without** any processing applied, meaning you will usually get a frame you need to decode yourself.
     /// # Errors
     /// If the backend fails to get the frame (e.g. already taken, busy, doesn't exist anymore), or [`open_stream()`](CaptureBackendTrait::open_stream()) has not been called yet, this will error.
-    pub fn frame_raw(&self) -> Result<Vec<u8>, NokhwaError> {
-        self.backend.borrow_mut().frame_raw()
+    pub fn frame_raw(&mut self) -> Result<Cow<[u8]>, NokhwaError> {
+        match self.backend.frame_raw() {
+            Ok(f) => Ok(f),
+            Err(why) => Err(why),
+        }
     }
 
     /// The minimum buffer size needed to write the current frame (RGB24). If `rgba` is true, it will instead return the minimum size of the RGBA buffer needed.
+    #[must_use]
     pub fn min_buffer_size(&self, rgba: bool) -> usize {
-        let resolution = self.backend.borrow().resolution();
+        let resolution = self.backend.resolution();
         if rgba {
             return (resolution.width() * resolution.height() * 4) as usize;
         }
@@ -187,19 +223,31 @@ impl Camera {
     /// # Errors
     /// If the backend fails to get the frame (e.g. already taken, busy, doesn't exist anymore), or [`open_stream()`](CaptureBackendTrait::open_stream()) has not been called yet, this will error.
     pub fn frame_to_buffer(
-        &self,
+        &mut self,
         buffer: &mut [u8],
         convert_rgba: bool,
     ) -> Result<usize, NokhwaError> {
-        let frame = self.frame()?;
-        let mut frame_data = frame.to_vec();
+        let resolution = self.resolution();
+        let frame = self.frame_raw()?;
         if convert_rgba {
-            let rgba_image: RgbaImage = frame.convert();
-            frame_data = rgba_image.to_vec();
+            let image_data =
+                match ImageBuffer::from_raw(resolution.width(), resolution.height(), frame) {
+                    Some(image) => {
+                        let image: ImageBuffer<Rgb<u8>, Cow<[u8]>> = image;
+                        image
+                    }
+                    None => {
+                        return Err(NokhwaError::ReadFrameError(
+                            "Frame Cow Too Small".to_string(),
+                        ))
+                    }
+                };
+            let rgba_image: RgbaImage = image_data.convert();
+            buffer.copy_from_slice(rgba_image.as_raw());
+            return Ok(rgba_image.len());
         }
-        let bytes = frame_data.len();
-        buffer.copy_from_slice(&frame_data);
-        Ok(bytes)
+        buffer.copy_from_slice(&frame);
+        Ok(frame.len())
     }
 
     #[cfg(feature = "output-wgpu")]
@@ -229,17 +277,17 @@ impl Camera {
             sample_count: 1,
             dimension: TextureDimension::D2,
             format: TextureFormat::Rgba8UnormSrgb,
-            usage: TextureUsage::SAMPLED | TextureUsage::COPY_DST,
+            usage: TextureUsages::TEXTURE_BINDING | TextureUsages::COPY_DST,
         });
 
         let width_nonzero = match NonZeroU32::try_from(4 * rgba_frame.width()) {
             Ok(w) => Some(w),
-            Err(why) => return Err(NokhwaError::CouldntCaptureFrame(why.to_string())),
+            Err(why) => return Err(NokhwaError::ReadFrameError(why.to_string())),
         };
 
         let height_nonzero = match NonZeroU32::try_from(rgba_frame.height()) {
             Ok(h) => Some(h),
-            Err(why) => return Err(NokhwaError::CouldntCaptureFrame(why.to_string())),
+            Err(why) => return Err(NokhwaError::ReadFrameError(why.to_string())),
         };
 
         queue.write_texture(
@@ -247,6 +295,7 @@ impl Camera {
                 texture: &texture,
                 mip_level: 0,
                 origin: wgpu::Origin3d::ZERO,
+                aspect: TextureAspect::All,
             },
             &rgba_frame.to_vec(),
             ImageDataLayout {
@@ -263,8 +312,8 @@ impl Camera {
     /// Will drop the stream.
     /// # Errors
     /// Please check the `Quirks` section of each backend.
-    pub fn stop_stream(&self) -> Result<(), NokhwaError> {
-        self.backend.borrow_mut().stop_stream()
+    pub fn stop_stream(&mut self) -> Result<(), NokhwaError> {
+        self.backend.stop_stream()
     }
 }
 
@@ -275,15 +324,16 @@ impl Drop for Camera {
 }
 
 // TODO: Update as we go
+#[allow(clippy::ifs_same_cond)]
 fn figure_out_auto() -> Option<CaptureAPIBackend> {
     let platform = std::env::consts::OS;
     let mut cap = CaptureAPIBackend::Auto;
     if cfg!(feature = "input-v4l") && platform == "linux" {
-        cap = CaptureAPIBackend::Video4Linux
+        cap = CaptureAPIBackend::Video4Linux;
     } else if cfg!(feature = "input-msmf") && platform == "windows" {
-        cap = CaptureAPIBackend::Windows
-    } else if cfg!(feature = "input-avfoundationn") && platform == "mac" {
-        cap = CaptureAPIBackend::AVFoundation
+        cap = CaptureAPIBackend::MediaFoundation;
+    } else if cfg!(feature = "input-avfoundation") && (platform == "macos" || platform == "ios") {
+        cap = CaptureAPIBackend::AVFoundation;
     } else if cfg!(feature = "input-uvc") {
         cap = CaptureAPIBackend::UniversalVideoClass;
     } else if cfg!(feature = "input-gst") {
@@ -299,11 +349,11 @@ fn figure_out_auto() -> Option<CaptureAPIBackend> {
 
 macro_rules! cap_impl_fn {
     {
-        $( ($backend:ty, $init_fn:ident, $feature:expr, $backend_name:ident) ),+
+        $( ($backend:expr, $init_fn:ident, $cfg:meta, $backend_name:ident) ),+
     } => {
         $(
             paste::paste! {
-                #[cfg(feature = $feature)]
+                #[cfg ($cfg) ]
                 fn [< init_ $backend_name>](idx: usize, setting: Option<CameraFormat>) -> Option<Result<Box<dyn CaptureBackendTrait>, NokhwaError>> {
                     use crate::backends::capture::$backend;
                     match <$backend>::$init_fn(idx, setting) {
@@ -311,7 +361,7 @@ macro_rules! cap_impl_fn {
                         Err(why) => Some(Err(why)),
                     }
                 }
-                #[cfg(not(feature = $feature))]
+                #[cfg(not( $cfg ))]
                 fn [< init_ $backend_name>](_idx: usize, _setting: Option<CameraFormat>) -> Option<Result<Box<dyn CaptureBackendTrait>, NokhwaError>> {
                     None
                 }
@@ -341,29 +391,29 @@ macro_rules! cap_impl_matches {
                                                 Err(why) => return Err(why),
                                             }
                                             None => {
-                                                return Err(NokhwaError::NotImplemented(
-                                                    "Platform requirements not satisfied.".to_string(),
+                                                return Err(NokhwaError::NotImplementedError(
+                                                    "Platform requirements not satisfied (Wrong Platform - Not Implemented).".to_string(),
                                                 ));
                                             }
                                         }
                                     }
                                     false => {
-                                        return Err(NokhwaError::NotImplemented(
-                                            "Platform requirements not satisfied.".to_string(),
+                                        return Err(NokhwaError::NotImplementedError(
+                                            "Platform requirements not satisfied. (Wrong Platform - Not Selected)".to_string(),
                                         ));
                                     }
                                 }
                             }
                         )+
                         _ => {
-                            return Err(NokhwaError::NotImplemented(
-                                "Platform requirements not satisfied.".to_string(),
+                            return Err(NokhwaError::NotImplementedError(
+                                "Platform requirements not satisfied. (Invalid Backend)".to_string(),
                             ));
                         }
                     }
                     None => {
-                        return Err(NokhwaError::NotImplemented(
-                            "Platform requirements not satisfied.".to_string(),
+                        return Err(NokhwaError::NotImplementedError(
+                            "Platform requirements not satisfied. (No Selection)".to_string(),
                         ));
                     }
                 }
@@ -377,35 +427,32 @@ macro_rules! cap_impl_matches {
                                         Err(why) => return Err(why),
                                     }
                                     None => {
-                                        return Err(NokhwaError::NotImplemented(
-                                            "Platform requirements not satisfied.".to_string(),
+                                        return Err(NokhwaError::NotImplementedError(
+                                            "Platform requirements not satisfied (Wrong Platform - Not Implemented).".to_string(),
                                         ));
                                     }
                                 }
                             }
                             false => {
-                                return Err(NokhwaError::NotImplemented(
-                                    "Platform requirements not satisfied.".to_string(),
+                                return Err(NokhwaError::NotImplementedError(
+                                    "Platform requirements not satisfied. (Wrong Platform - Not Selected)".to_string(),
                                 ));
                             }
                         }
                     }
                 )+
-                _ => {
-                    return Err(NokhwaError::NotImplemented(
-                        "Platform requirements not satisfied.".to_string(),
-                    ));
-                }
             }
-        };
+        }
     }
 }
 
 cap_impl_fn! {
-    (GStreamerCaptureDevice, new, "input-gst", gst),
-    (OpenCvCaptureDevice, new_autopref, "input-opencv", opencv),
-    (V4LCaptureDevice, new, "input-v4l", v4l),
-    (UVCCaptureDevice, create, "input-uvc", uvc)
+    (GStreamerCaptureDevice, new, feature = "input-gst", gst),
+    (OpenCvCaptureDevice, new_autopref, feature = "input-opencv", opencv),
+    (UVCCaptureDevice, create, feature = "input-uvc", uvc),
+    (V4LCaptureDevice, new, all(feature = "input-v4l", target_os = "linux"), v4l),
+    (MediaFoundationCaptureDevice, new, all(feature = "input-msmf", target_os = "windows"), msmf),
+    (AVFoundationCaptureDevice, new, all(feature = "input-avfoundation", any(target_os = "macos", target_os = "ios")), avfoundation)
 }
 
 fn init_camera(
@@ -416,9 +463,14 @@ fn init_camera(
     let camera_backend = cap_impl_matches! {
             backend, index, format,
             ("input-v4l", Video4Linux, init_v4l),
+            ("input-msmf", MediaFoundation, init_msmf),
+            ("input-avfoundation", AVFoundation, init_avfoundation),
             ("input-uvc", UniversalVideoClass, init_uvc),
             ("input-gst", GStreamer, init_gst),
             ("input-opencv", OpenCv, init_opencv)
     };
     Ok(camera_backend)
 }
+
+#[cfg(feature = "output-threaded")]
+unsafe impl Send for Camera {}
